from typing import Union


def rgb_to_int(rgb: Union[tuple, dict, int]) -> int:
    if isinstance(rgb, int):
        return rgb
    elif isinstance(rgb, tuple):
        red, green, blue = rgb
    elif isinstance(rgb, dict):
        red = rgb['red']
        green = rgb['green']
        blue = rgb['blue']
    else:
        raise ValueError("Invalid value for RGB!")

    r = red << 16
    g = green << 8
    b = blue

    return r+g+b


def int_to_rgb(rgb: int) -> tuple:
    red = (rgb & 16711680) >> 16
    green = (rgb & 65280) >> 8
    blue = (rgb & 255)
    return red, green, blue





from datetime import datetime


class BatteryInfo(object):
    def __init__(self, battery_charge: float, sample_ts: datetime):
        self._battery_charge = battery_charge
        self._sample_ts = sample_ts

    @property
    def remaining_charge(self) -> float:
        """
        Expresses the remaining battery charge in percentage

        :return:
        """
        return self._battery_charge

    @property
    def sampled_datetime(self) -> datetime:
        """
        When the battery info has been sampled (timestamp dependant on your current timezone)
        :return:
        """
        return self._sample_ts



import logging
from datetime import datetime
from typing import Union, List

from meross_iot.model.enums import OnlineStatus
from meross_iot.model.shared import BaseDictPayload

_LOGGER = logging.getLogger(__name__)


class HttpDeviceInfo(BaseDictPayload):
    def __init__(self,
                 uuid: str,
                 online_status: Union[int, OnlineStatus],
                 dev_name: str,
                 dev_icon_id: str,
                 bind_time: Union[int, datetime],
                 device_type: str,
                 sub_type: str,
                 channels: List[dict],
                 region: str,
                 fmware_version: str,
                 hdware_version: str,
                 user_dev_icon: str,
                 icon_type: int,
                 skill_number: str,
                 domain: str,
                 reserved_domain: str,
                 *args, **kwargs):

        super().__init__(*args, **kwargs)
        self.uuid = uuid
        if isinstance(online_status, int):
            self.online_status = OnlineStatus(online_status)
        elif isinstance(online_status, OnlineStatus):
            self.online_status = online_status
        else:
            _LOGGER.warning(f"Provided online_status is not int neither OnlineStatus. It will be ignored.")
            self.online_status = None

        self.dev_name = dev_name
        self.dev_icon_id = dev_icon_id
        if isinstance(bind_time, int):
            self.bind_time = datetime.utcfromtimestamp(bind_time)
        elif isinstance(bind_time, datetime):
            self.bind_time = bind_time
        else:
            _LOGGER.warning(f"Provided bind_time is not int neither datetime. It will be ignored.")
            self.bind_time = None

        self.device_type = device_type
        self.sub_type = sub_type
        self.channels = channels
        self.region = region
        self.fmware_version = fmware_version
        self.hdware_version = hdware_version
        self.user_dev_icon = user_dev_icon
        self.icon_type = icon_type
        self.skill_number = skill_number
        self.domain = domain
        self.reserved_domain = reserved_domain




import logging
from enum import Enum
from typing import Union


_LOGGER = logging.getLogger(__name__)


class OnlineStatus(Enum):

    NOT_ONLINE = 0
    ONLINE = 1
    OFFLINE = 2
    UNKNOWN = -1
    UPGRADING = 3


class LightMode(Enum):
    MODE_LUMINANCE = 4
    MODE_TEMPERATURE = 2
    MODE_RGB = 1
    MODE_RGB_LUMINANCE = 5
    MODE_TEMPERATURE_LUMINANCE = 6


class SprayMode(Enum):
    OFF = 0
    CONTINUOUS = 1
    INTERMITTENT = 2


class ThermostatV3Mode(Enum):
    AUTO = 3
    COOL = 2
    CUSTOM = 0
    ECONOMY = 4
    HEAT = 1


class Namespace(Enum):
    # Common abilities
    SYSTEM_ALL = 'Appliance.System.All'
    SYSTEM_ABILITY = 'Appliance.System.Ability'
    SYSTEM_ONLINE = 'Appliance.System.Online'
    SYSTEM_REPORT = 'Appliance.System.Report'
    SYSTEM_DEBUG = 'Appliance.System.Debug'

    CONTROL_BIND = 'Appliance.Control.Bind'
    CONTROL_UNBIND = 'Appliance.Control.Unbind'
    CONTROL_TRIGGER = 'Appliance.Control.Trigger'
    CONTROL_TRIGGERX = 'Appliance.Control.TriggerX'

    CONFIG_WIFI_LIST = 'Appliance.Config.WifiList'
    CONFIG_TRACE = 'Appliance.Config.Trace'

    # Power plug/bulbs abilities
    CONTROL_TOGGLE = 'Appliance.Control.Toggle'
    CONTROL_TOGGLEX = 'Appliance.Control.ToggleX'
    CONTROL_ELECTRICITY = 'Appliance.Control.Electricity'
    CONTROL_CONSUMPTIONX = 'Appliance.Control.ConsumptionX'

    # Bulbs-only abilities
    CONTROL_LIGHT = 'Appliance.Control.Light'

    # Garage opener abilities
    GARAGE_DOOR_STATE = 'Appliance.GarageDoor.State'

    # Humidifier
    CONTROL_SPRAY = 'Appliance.Control.Spray'

    SYSTEM_DIGEST_HUB = 'Appliance.Digest.Hub'

    # HUB
    HUB_EXCEPTION = 'Appliance.Hub.Exception'
    HUB_BATTERY = 'Appliance.Hub.Battery'
    HUB_TOGGLEX = 'Appliance.Hub.ToggleX'
    HUB_ONLINE = 'Appliance.Hub.Online'

    # SENSORS
    HUB_SENSOR_ALL = 'Appliance.Hub.Sensor.All'
    HUB_SENSOR_TEMPHUM = 'Appliance.Hub.Sensor.TempHum'
    HUB_SENSOR_ALERT = 'Appliance.Hub.Sensor.Alert'

    # MTS100
    HUB_MTS100_ALL = 'Appliance.Hub.Mts100.All'
    HUB_MTS100_TEMPERATURE = 'Appliance.Hub.Mts100.Temperature'
    HUB_MTS100_MODE = 'Appliance.Hub.Mts100.Mode'


def get_or_parse_namespace(namespace: Union[Namespace, str]):
    if isinstance(namespace, str):
        try:
            parsed_namespace = Namespace(namespace)
            return parsed_namespace
        except ValueError:
            _LOGGER.error(f"Namespace {namespace} is not currently handled/recognized.")
            raise
    elif isinstance(namespace, Namespace):
        return namespace
    else:
        raise ValueError("Unknown invalid namespace type. Only str/Namespace types are allowed here.")


from meross_iot.model.enums import Namespace
from meross_iot.model.push.generic import GenericPushNotification


class UnbindPushNotification(GenericPushNotification):
    def __init__(self, originating_device_uuid: str, raw_data: dict):
        super().__init__(namespace=Namespace.CONTROL_UNBIND, originating_device_uuid=originating_device_uuid, raw_data=raw_data)

    @classmethod
    def from_dict(cls, data: dict, originating_device_uuid: str):
        return UnbindPushNotification(originating_device_uuid=originating_device_uuid, raw_data=data)



import uuid as UUID
from hashlib import md5

def build_device_request_topic(client_uuid: str) -> str:
    """
    Builds the MQTT topic where commands should be send to specific devices
    :param client_uuid:
    :return:
    """
    return f"/appliance/{client_uuid}/subscribe"


def build_client_response_topic(user_id: str, app_id: str) -> str:
    """
    Builds the MQTT topic where the device sends back ACKs to commands
    :param app_id:
    :param user_id:
    :param client_uuid:
    :return:
    """
    return f"/app/{user_id}-{app_id}/subscribe"


def build_client_user_topic(user_id: str):
    """
    Builds the topic name where user push notification are received
    :param user_id:
    :return:
    """
    return f"/app/{user_id}/subscribe"


def device_uuid_from_push_notification(from_topic: str):
    """
    Extracts the device uuid from the "from" header of the received messages.
    :param from_topic:
    :return:
    """
    return from_topic.split('/')[2]


def generate_client_and_app_id():
    """
    Generates a new app-id.
    :return:
    """
    # TODO: Talk to the Meross engineer and check if the APPID should be the same or if we
    #  need to use a convention to discriminate MerossIot python clients.
    md5_hash = md5()
    rnd_uuid = UUID.uuid4()
    md5_hash.update(f"API{rnd_uuid}".encode("utf8"))
    app_id = md5_hash.hexdigest()
    client_id = 'app:%s' % md5_hash.hexdigest()
    return app_id, client_id


def generate_mqtt_password(user_id: str, key: str):
    """
    Generates the MQTT password that the APP uses to connect to the mqtt server.
    :param user_id:
    :param key:
    :return:
    """
    md5_hash = md5()
    clearpwd = f"{user_id}{key}"
    md5_hash.update(clearpwd.encode("utf8"))
    return md5_hash.hexdigest()


def verify_message_signature(header: dict, key: str):
    """
    Verifies if the given message header has a valid signature
    :param header:
    :param key:
    :return:
    """
    message_hash = md5()
    strtohash = "%s%s%s" % (header['messageId'], key, header['timestamp'])
    message_hash.update(strtohash.encode("utf8"))
    expected_signature = message_hash.hexdigest().lower()
    return expected_signature == header['sign']
